/*
 * nftw() function, (hopefully) as described in nftw(XSH), XPG 6.
 *
 * Gunnar Ritter, Freiburg i. Br., Germany, July 2002.
 */
/*
 * Copyright (c) 2003 Gunnar Ritter
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute
 * it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 */

/*	Sccsid @(#)nftw.c	1.5 (gritter) 4/3/04	*/

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<limits.h>
#include	<errno.h>
#include	<string.h>
#include	<stdlib.h>
#include	<dirent.h>

#include	"ftw.h"

#include	"getdir.h"

struct	visited {
	ino_t	v_ino;
	dev_t	v_dev;
};

struct	olddir {
	char	*o_name;
	int	o_fd;
};

static int
nftw_internal(const char	*path,
	const char	*base,	/* last component of path */
	int	(*fn)(const char *, const struct stat *, int, struct FTW *),
	int	curdepth,
	const int	maxdepth,
	const int	flags,
	struct FTW	*ftwp,
	int	(*statf)(const char *, struct stat *),	/* stat or lstat */
	struct olddir	*odir,
	struct visited	*vp)
{
	struct getdb	*db;
	struct direc	*dp;
	int	df, err;
	struct stat	st, pst;
	char	*file, *filend;
	int	val = 0;
	size_t	sz, slen, ss;

	if (++curdepth > maxdepth)
		goto end0;
	if ((df = open((flags & FTW_CHDIR) ? base : path, O_RDONLY)) < 0 ||
	    (db = getdb_alloc((flags & FTW_CHDIR) ? base : path, df)) == NULL) {
		if (errno == EACCES) {
			stat((flags & FTW_CHDIR) ? base : path, &pst);
			val = fn(path, &pst, FTW_DNR, ftwp);
		} else
			val = -1;
		goto end0;
	}
	if (fstat(df, &pst) < 0 || (flags & FTW_CHDIR) && fchdir(df) < 0) {
		if (errno == EACCES)
			val = fn(path, &pst, FTW_DNR, ftwp);
		else
			val = -1;
		goto end1;
	}
#ifndef	__dietlibc__
	if ((sz = fpathconf(df, _PC_PATH_MAX)) <= 0)
		sz = _POSIX_PATH_MAX;
#else	/* __dietlibc__ */
	sz = PATH_MAX;
#endif	/* __dietlibc__ */
	sz += ftwp->base + 1;
	if ((file = malloc(sz)) == NULL) {
		val = -1;
		goto end2;
	}
	filend = file;
	if (path[0] == '/' && path[1] == '\0')
		*filend++ = '/';
	else {
		const char	*cp;

		cp = path;
		while (*filend++ = *cp++);
		filend[-1] = '/';
	}
	slen = filend - file;
	if ((flags & FTW_DEPTH) == 0)
		if ((val = fn(path, &pst, FTW_D, ftwp)) != 0)
			goto end3;
	/*
	 * See <http://www.opengroup.org/platform/resolutions/bwg97-003.html>
	 * for more information about this.
	 */
	if ((flags & FTW_PHYS) == 0) {
		unsigned	i;

		for (i = 0; i < curdepth - 1; i++)
			if (vp[i].v_dev == pst.st_dev &&
					vp[i].v_ino == pst.st_ino)
				goto end3;
		vp[i].v_dev = pst.st_dev;
		vp[i].v_ino = pst.st_ino;
	}
	ftwp->level++;
	while ((dp = getdir(db, &err)) != NULL) {
		int	type;

		if (dp->d_name[0] == '.' && (dp->d_name[1] == '\0' ||
				dp->d_name[1] == '.' && dp->d_name[2] == '\0'))
			continue;
		if (slen + (ss = strlen(dp->d_name)) >= sz) {
			sz = slen + ss + 1;
			if ((file = realloc(file, sz)) == NULL) {
				ftwp->level--;
				goto end3;
			}
			filend = &file[slen];
		}
		strcpy(filend, dp->d_name);
		if (statf((flags & FTW_CHDIR) ? dp->d_name : file, &st) >= 0) {
			if (flags & FTW_MOUNT)
				if (st.st_dev != pst.st_dev)
					continue;
			if ((flags & FTW_PHYS) && S_ISLNK(st.st_mode)) {
				type = FTW_SL;
			} else if (S_ISDIR(st.st_mode))
				type = FTW_D;
			else
				type = FTW_F;
		} else {
			struct stat	lst;

			if ((flags & FTW_PHYS) == 0 && lstat(file, &lst) == 0)
				type = FTW_SLN;
			else if (errno == EACCES)
				type = FTW_NS;
			else {
				val = -1;
				ftwp->level--;
				goto end3;
			}
		}
		if (type == FTW_D) {
			struct olddir	od;

			od.o_name = NULL;
			od.o_fd = df;
			ftwp->base += ss;
			val = nftw_internal(file, dp->d_name, fn,
					curdepth, maxdepth,
					flags, ftwp, statf, &od, vp);
			ftwp->base -= ss;
		} else
			val = fn(file, &st, type, ftwp);
		if (val != 0) {
			ftwp->level--;
			goto end3;
		}
	}
	ftwp->level--;
	if (flags & FTW_DEPTH)
		val = fn(path, &pst, FTW_DP, ftwp);
end3:
	free(file);
end2:
	if (flags & FTW_CHDIR)
		if ((odir->o_fd < 0 ? chdir(odir->o_name) :
					fchdir(odir->o_fd)) < 0)
			val = -1;
end1:
	getdb_free(db);
	close(df);
end0:
	return val;
}

int
nftw(const char	*path,
	int	(*fn)(const char *, const struct stat *, int, struct FTW *),
	int	depth,
	int	flags)
{
	struct FTW	sftw;
	struct visited	*vp = NULL;
	int	(*statf)(const char *, struct stat *)
			= flags & FTW_PHYS ? lstat : stat;
	int	val;
	struct olddir	curdir;

	if (depth <= 0)
		return 0;
	sftw.level = 0;
	sftw.base = strlen(path) + 1;
	if (flags & FTW_CHDIR) {
		/*
		 * Get a file descriptor of the starting directory so one
		 * can change back with fchdir() later. This has two
		 * advantages in comparison to chdir(): First, we don't
		 * need a getcwd() call which may give bogus results;
		 * second, we return to the starting directory even if
		 * it has been moved.
		 *
		 * This might violate XPG 6 ("At most one file descriptor
		 * shall be used for each directory level"), but that
		 * seems acceptable as it only affects the lowest level.
		 *
		 * If the directory is not readable, try getcwd() as it
		 * may be executable.
		 */
		curdir.o_name = NULL;
		if ((curdir.o_fd = open(".", O_RDONLY)) < 0 ||
				fcntl(curdir.o_fd, F_SETFD, FD_CLOEXEC) < 0) {
			size_t len = 0;
			for (;;) {
				if ((curdir.o_name = realloc(curdir.o_name,
						len += 256)) != NULL) {
					if (getcwd(curdir.o_name, len) != NULL)
						break;
					if (errno == ERANGE)
						continue;
				}
				val = -1;
				goto end0;
			}
		}
	}
	if ((flags & FTW_PHYS) == 0) {
		int	cnt;

		if (depth < (cnt = sysconf(_SC_OPEN_MAX)))
			cnt = depth;
		if ((vp = malloc(sizeof *vp * cnt)) == NULL) {
			val = -1;
			goto end1;
		}
	}
	val = nftw_internal(path, path, fn,
			0, depth, flags, &sftw, statf, &curdir, vp);
	if ((flags & FTW_PHYS) == 0)
		free(vp);
end1:
	if (flags & FTW_CHDIR) {
		if (curdir.o_fd >= 0)
			close(curdir.o_fd);
		if (curdir.o_name != NULL)
			free(curdir.o_name);
	}
end0:
	return val;
}
