/*
 * wc - word count
 *
 * Gunnar Ritter, Freiburg i. Br., Germany, December 2000.
 */
/*
 * Copyright (c) 2003 Gunnar Ritter
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute
 * it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 */

#if __GNUC__ >= 3 && __GNUC_MINOR__ >= 4
#define	USED	__attribute__ ((used))
#elif defined __GNUC__
#define	USED	__attribute__ ((unused))
#else
#define	USED
#endif
#if defined (S42)
static const char sccsid[] USED = "@(#)wc_s42.sl	1.32 (gritter) 4/20/04";
#elif defined (SUS)
static const char sccsid[] USED = "@(#)wc_sus.sl	1.32 (gritter) 4/20/04";
#else
static const char sccsid[] USED = "@(#)wc.sl	1.32 (gritter) 4/20/04";
#endif

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<wchar.h>
#include	<locale.h>
#include	<ctype.h>
#include	<wctype.h>
#include	<unistd.h>
#include	<libgen.h>
#include	<limits.h>

#include	<iblok.h>
#include	<blank.h>

static unsigned		errcnt;		/* count of errors */
static int		cflag;		/* count bytes only */
static int		lflag;		/* count lines only */
static int		mflag;		/* count characters only */
static int		wflag;		/* count words only */
static off_t		bytec;		/* byte count */
static off_t		charc;		/* character count */
static off_t		wordc;		/* word count */
static off_t		linec;		/* line count */
static off_t		chart;		/* total character count */
static off_t		bytet;		/* total byte count */
static off_t		wordt;		/* total word count */
static off_t		linet;		/* total line count */
#if defined (SUS) || defined (S42)
static int		putspace;	/* wrote space to output line */
#endif
static char		*progname;	/* argv0 to main */
static int		multibyte;	/* MB_CUR_MAX > 1 */

/*
 * Format output.
 */
static void
report(unsigned long long count)
{
#if defined (SUS) || defined (S42)
	if (putspace++)
		printf(" ");
	printf("%llu", count);
#else	/* !SUS, !S42 */
	printf("%7llu ", count);
#endif	/* !SUS, !S42 */
}

static int
sbwc(struct iblok *ip)
{
	register off_t hadspace = 1;
	register int c, i;
	size_t sz;

	while (ib_read(ip) != EOF) {
		ip->ib_cur--;
		sz = ip->ib_end - ip->ib_cur;
		charc += sz;
		for (i = 0; i < sz; i++) {
			c = ip->ib_cur[i] & 0377;
#ifdef	SUS
			if (isspace(c)) {
				if (c == '\n')
					linec++;
				hadspace++;
			} else {
				if (hadspace) {
					hadspace = 0;
					wordc++;
				}
			}
#else	/* !SUS */
			if (c == '\n') {
				linec++;
				hadspace++;
			} else if (isblank(c)) {
				hadspace++;
			} else {
				if (hadspace && isgraph(c)) {
					hadspace = 0;
					wordc++;
				}
			}
#endif	/* !SUS */
		}
	}
	bytec = charc;
	return 0;
}

static int
mbwc(struct iblok *ip)
{
	off_t	hadspace = 1;
	char	*cp;
	wchar_t	wc;
	int	n;

	while ((cp = ib_getw(ip, &wc, &n)) != NULL) {
		bytec += n;
		if (wc != WEOF) {
			charc++;
#ifdef	SUS
			if (iswspace(wc)) {
				if (wc == '\n')
					linec++;
				hadspace++;
			} else {
				if (hadspace) {
					hadspace = 0;
					wordc++;
				}
			}
#else	/* !SUS */
			if (wc == '\n') {
				linec++;
				hadspace++;
			} else if (iswblank(wc)) {
				hadspace++;
			} else {
				if (hadspace && iswgraph(wc)) {
					hadspace = 0;
					wordc++;
				}
			}
#endif	/* !SUS */
		}
	}
	return 0;
}

/*
 * Count a file.
 */
static int
fpwc(int fd)
{
	struct iblok	*ip;

	bytec = charc = wordc = linec = 0;
	ip = ib_alloc(fd, 0);
	if ((multibyte ? mbwc : sbwc)(ip)) {
		ib_free(ip);
		return -1;
	}
	ib_free(ip);
#if defined (SUS) || defined (S42)
	putspace = 0;
#endif
	if (lflag)
		report(linec);
	if (wflag)
		report(wordc);
	if (cflag)
		report(bytec);
	if (mflag)
		report(charc);
	linet += linec;
	wordt += wordc;
	chart += charc;
	bytet += bytec;
	return 0;
}

/*
 * Count a file given by its name.
 */
static int
filewc(const char *fn)
{
	int fd;
	int r;

	if ((fd = open(fn, O_RDONLY)) < 0) {
		fprintf(stderr, "%s: cannot open %s\n", progname, fn);
		errcnt |= 1;
		return 0;
	}
	if ((r = fpwc(fd)) >= 0)
		printf(" %s\n", fn);
	if (fd > 0)
		close(fd);
	return r >= 0 ? 1 : 0;
}

static void
usage(void)
{
	fprintf(stderr, "usage: %s [-cmlw] [name ...]\n", progname);
	exit(2);
}

int
main(int argc, char **argv)
{
	int c;
	unsigned ac;

#ifdef	__GLIBC__
	putenv("POSIXLY_CORRECT=1");
#endif
	progname = basename(argv[0]);
	setlocale(LC_CTYPE, "");
	multibyte = MB_CUR_MAX > 1;
	ac = 0;
	while ((c = getopt(argc, argv, ":clmw")) != EOF) {
		ac++;
		switch (c) {
		case 'c':
			cflag = ac;
			break;
		case 'l':
			lflag = ac;
			break;
		case 'm':
			mflag = ac;
			break;
		case 'w':
			wflag = ac;
			break;
		default:
			usage();
		}
	}
	if (ac == 0) {
#if !defined (SUS) && !defined(S42)
		if (argv[1] && argv[1][0] == '-' && argv[1][1] == '\0')
			optind++;
		else
#endif	/* !SUS, !S42 */
			cflag = lflag = wflag = 1;
	}
	if (mflag)
		cflag = 0;
	if (optind < argc) {
		ac = 0;
		while (optind < argc)
			ac += filewc(argv[optind++]);
		if (ac > 1) {
#if defined (SUS) || defined (S42)
			putspace = 0;
#endif
			if (lflag)
				report(linet);
			if (wflag)
				report(wordt);
			if (cflag)
				report(bytet);
			if (mflag)
				report(chart);
			printf(" total\n");
		}
	} else {
		fpwc(0);
		printf("\n");
	}
	return errcnt;
}
